import java.util.Arrays;

/**
 * Classe que representa o estado do tabuleiro de um jogo que eh uma versao 
 * simples da batalha naval (so com submarinos)
 * 
 * @author Ana Santos, Antonia Lopes
 * @version 1.0 15/Jul/2013
 */

public class EstadoTabuleiro implements Cloneable {
	
	/**
	 * Enumerado que representa o estado das casas do tabuleiro
	 */
	public enum Estado {
	    LIVRE_ESCONDIDA, OCUPADA_ESCONDIDA, LIVRE_REVELADA, OCUPADA_REVELADA;
	   
	   /**
	    * Representacao textual dos elementos do enumerado
	    * @return String com representacao textual do enumerado 
	    */
	    @Override
	    public String toString() {
	    	if (this == LIVRE_ESCONDIDA) return "   ";
	    	if (this == OCUPADA_ESCONDIDA) return " x ";
	    	if (this == LIVRE_REVELADA) return " - ";
	    	if (this == OCUPADA_REVELADA) return " o ";
			return null;
	    };
	}

	/**
	 * Matriz que representa estado do tabuleiro
	 */
	private Estado[][] tabuleiro;
	
	/**
	 * Construtor que constroi um tabuleiro quadrado com dim linhas e dim colunas 
	 * com todas as casas todas desocupadas e nao reveladas
	 * @param dim - dimensao do tabuleiro
	 * @requires dim >= 1
	 */
	public EstadoTabuleiro(int dim) {
		
		tabuleiro = new Estado[dim][dim];
		for (int i = 0; i < dim; i++) 
			for (int j = 0; j < dim; j++)
				tabuleiro[i][j] = Estado.LIVRE_ESCONDIDA;
	}
	
	/**
	 * Devolve a dimensao do tabuleiro;
	 * @return dimensao do tabuleiro
	 * @ensures dimensao() >=1 
	 */
	public int dimensao() {
		return tabuleiro.length;
	}
	
	/**
	 * Determina se a casa na linha e coluna indicadas esta ocupada
	 * @param linha - linha pretendida
	 * @param coluna - coluna pretendida
	 * @return - true se casa estiver ocupada, false caso contrario
	 * @requires linha >= 1  &&  coluna >= 1 
	 * 			 && linha <= dimensao() && coluna <= dimensao() 
	 */
	public boolean ocupada(int linha, int coluna)  {
		return tabuleiro[linha-1][coluna-1].equals(Estado.OCUPADA_ESCONDIDA) 
		       || tabuleiro[linha-1][coluna-1].equals(Estado.OCUPADA_REVELADA);
	}
	
	/**
	 * Determina se a casa na linha e coluna indicadas esta revelada
	 * @param linha - linha pretendida
	 * @param coluna - coluna pretendida
	 * @return - true se casa estiver revelada, false caso contrario
	 * @requires linha >= 1  &&  coluna >= 1 
	 * 			&& linha <= dimensao() && coluna <= dimensao() 
	 */
	public boolean revelada(int linha, int coluna)  {
		return tabuleiro[linha-1][coluna-1].equals(Estado.LIVRE_REVELADA) 
		       || tabuleiro[linha-1][coluna-1].equals(Estado.OCUPADA_REVELADA);
	}

	
	/**
	 * Torna a casa na linha e coluna indicadas ocupada e nao revelada;
	 * @param linha - linha pretendida
	 * @param coluna - coluna pretendida
	 * @requires linha >= 1  &&  coluna >= 1 
	 * 			 && linha <= dimensao() && coluna <= dimensao() 	 
	 */
	public void ocupa(int linha, int coluna) {
		tabuleiro[linha-1][coluna-1] = Estado.OCUPADA_ESCONDIDA;
	}
	
	/**
	 * Determina se a casa na linha e coluna indicadas esta desocupada 
	 * e todas as casas contiguas estao tambem desocupadas 
	 * @param linha - linha pretendida
	 * @param coluna - coluna pretendida
	 * @return - true se a casa e contiguas estiverem livres, 
	 * 			false caso contrario
	 * @requires linha >= 1  &&  coluna >= 1 
	 * 			 && linha <= dimensao() && coluna <= dimensao() 
	 */
	public boolean livre(int linha, int coluna) {	
		
		boolean ocupada = ocupada(linha, coluna);
		if (!ocupada && linha > 1)
			ocupada = ocupada (linha-1, coluna);
		if (!ocupada && coluna > 1)
			ocupada = ocupada (linha, coluna-1);
		if (!ocupada && linha < dimensao())
			ocupada = ocupada (linha+1, coluna);
		if (!ocupada && coluna < dimensao())
			ocupada = ocupada (linha, coluna+1);
		if (!ocupada && linha > 1 && coluna > 1)
			ocupada = ocupada (linha-1, coluna-1);
		if (!ocupada && linha < dimensao() && coluna < dimensao())
			ocupada = ocupada (linha+1, coluna+1);
		if (!ocupada && linha > 1 && coluna < dimensao())
			ocupada = ocupada (linha-1, coluna+1);
		if (!ocupada && linha < dimensao() && coluna > 1)
			ocupada = ocupada (linha+1, coluna-1);
		return !ocupada;
	}
	
	/**
	 * Ttorna a casa na linha e coluna indicadas revelada;
	 * @param linha - linha pretendida
	 * @param coluna - coluna pretendida
	 * @requires linha >= 1  &&  coluna >= 1 
	 * 			&& linha <= dimensao() && coluna <= dimensao() 
	 */
	public void revela(int linha, int coluna) {
		if (tabuleiro[linha-1][coluna-1].equals(Estado.LIVRE_ESCONDIDA)) 
			tabuleiro[linha-1][coluna-1] = Estado.LIVRE_REVELADA;
		if (tabuleiro[linha-1][coluna-1].equals(Estado.OCUPADA_ESCONDIDA)) 
			tabuleiro[linha-1][coluna-1] = Estado.OCUPADA_REVELADA;
	}
	
	/**
	 * Indica se todas as casas ocupadas estao reveladas.
	 * @return True se todas as casas ocupadas estao reveladas, 
	 * 		   false caso contrario
	 */
	public boolean revelado() {
		boolean ret = true;
		for (int i = 0; i < tabuleiro.length; i++) 
			for (int j = 0; j < tabuleiro[i].length; j++)
				if (tabuleiro[i][j].equals(Estado.OCUPADA_ESCONDIDA))
					return false;
		return ret;
	}
	
	/**
	 * Devolve copia do EstadoTabuleiro
	 * @return clone do EstadoTabuleiro
	 */
	@Override
	public EstadoTabuleiro clone(){
		EstadoTabuleiro clone = null;
		try {
			clone = (EstadoTabuleiro) super.clone();
			System.arraycopy(tabuleiro, 0, clone.tabuleiro, 0, tabuleiro.length);
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return clone;
	}

	/**
	 * Devolve o hash code deste EstadoTabuleiro
	 * @return hash code
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(tabuleiro);
		return result;
	}

	/**
	 * Verifica se o Object dado eh igual a este EstadoTabuleiro
	 * @param obj - objeto a comparar
	 * @return true se os EstadoTabuleiro forem iguais, false caso contrario
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (obj instanceof EstadoTabuleiro)
			return false;
		EstadoTabuleiro other = (EstadoTabuleiro) obj;
		if (!Arrays.equals(tabuleiro, other.tabuleiro))
			return false;
		return true;
	}
	
	/**
	 * Devolve a representacao textual do estado do tabuleiro
	 * @return String que representa estado do tabuleiro
	 */
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder
				("\n   1  2  3  4  5  6  7  8 \n");
		for (int i = 0; i < this.dimensao(); i++) {
			result.append((i+1) + " ");
			for (int j = 0; j < this.dimensao(); j++)
				result.append(tabuleiro[i][j]);
			result.append('\n');
		}
		result.append('\n');
		return result.toString();
	}
}
