/**
 * Uma classe cujos objectos representam partidas de um jogo que eh 
 * uma versao simples da batalha naval (so com submarinos)
 * 
 * @author Antonia Lopes
 * @version 1.0 15/Jul/2013
 */
public class Ipex {

	/**
	 * Dimensao do tabuleiro em que se joga o jogo
	 * (value is 8)
	 */
	private static final int DIMENSAO = 8; 

	/**
	 * Numero de submarinos escondidos (por jogador)
	 * (value is 8)
	 */
	private static final int SUBMARINOS = 8; 

	/**
	 * Numero de tiros por jogada 
	 * (value is 3)
	 */
	private static final int TIROS = 3;

	/**
	 * Numero de jogadores que jogam o jogo
	 * (value is 2) 
	 */
	private static final int NUMERO_JOGADORES = 2; 

	// 	Atributos

	/**
	 * Tabuleiros do jogo - um por jogador
	 */
	private final EstadoTabuleiro[] tabuleiro = 
		new EstadoTabuleiro[NUMERO_JOGADORES];
	
	/**
	 * Jogador que tem a vez - chamado o jogador corrente (0 ou 1)
	 */
	private int jogadorVez;

	/**
	 * O numero da jogada em que o jogo vai
	 */
	private int numJogada = 1;
	
	/**
	 * O numero de tiros disparados pelo jogador corrente na jogada corrente
	 */
	private int tirosDisparados = 0;
	
	/**
	 * O numero de submarinos ja escondidos pelo jogador corrente 
	 * durante a primeira jogada
	 */
	private int submarinosEscondidos = 0;

	// 	Construtores

	/**
	 * Constroi novo jogo com tabuleiros de 8 casas
	 * @ensures dimensao() == 8 && jogador() == 1
	 */
	public Ipex(){	
		this.jogadorVez = 0;
		this.tabuleiro[0] = new EstadoTabuleiro(DIMENSAO);	
		this.tabuleiro[1] = new EstadoTabuleiro(DIMENSAO);	
	}

	// 	Metodos

	/** 
	 * Devolve a dimensao do Tabuleiro
	 * @return a dimensao do tabuleiro 
	 * */	
	public int dimensao(){	
		return DIMENSAO;
	}

	/** 
	 * Devolve o numero do jogador corrente
	 * @return o numero do jogador que tem a vez para jogar;
	 * @ensures \result == 1 || \result == 2
	 */
	public int jogador(){	
		return this.jogadorVez + 1;
	}

	/** 
	 * Devolve o numero da jogada corrente
	 * @return o numero da jogada corrente;
	 * @ensures \result >= 1
	 */
	public int jogada(){	
		return this.numJogada;
	}

	/** 
	 * Verifica se na casa dada pode ser guardado um submarino
	 * (casa esta livre, assim como as casas contiguas)
	 * @param linha - numero da linha
	 * @param coluna - numero da coluna
	 * @return true se na casa dada pode ser guardado um submarino,
	 * 		   caso contrario falso
	 */	
	private boolean cabeSubmarino(int linha, int coluna) {
		return this.tabuleiro[this.jogadorVez].livre(linha, coluna);
	}

	/** 
	 * Verifica se a jogada eh valida. 
	 * @param linha - numero da linha
	 * @param coluna - numero da coluna
	 * @return true se os numeros dados definem linha e coluna
	 * 		   do tabuleiro, caso contrario falso 
	 */	
	public boolean jogadaValida(int linha, int coluna) {
		
		if (!(linha >= 1 &&  coluna >= 1 
				&& linha <= DIMENSAO && coluna <= DIMENSAO)){
			return false;
		}
		if (this.numJogada > 1){
			return true;
		}
		return cabeSubmarino(linha, coluna);
	}

	/** 
	 * Efectua jogada
	 * @param linha - numero da linha
	 * @param coluna - numero da coluna
	 * @requires jogadaValida(linha, coluna) && !terminado()
	 */	
	public void efectuaJogada(int linha, int coluna) {
		if (this.numJogada == 1)
			escondeSubmarino(linha, coluna);
		else 
			daTiro(linha, coluna);
	}


	/** 
	 * Esconde submarino no tabuleiro do jogador corrente
	 * na casa dada
	 * @param linha - numero da linha
	 * @param coluna - numero da coluna
	 * @requires jogadaValida(linha, coluna) && jogada() = 1
	 */	
	private void escondeSubmarino(int linha, int coluna) {
		this.tabuleiro[this.jogadorVez].ocupa(linha, coluna);
		this.submarinosEscondidos++;
		if (this.submarinosEscondidos == SUBMARINOS){
			passaVez();
			this.submarinosEscondidos = 0;
		}
	}

	/**
	 * Da um tiro no tabuleiro do adversario do jogador corrente
	 * na casa dada
	 * @param linha - numero da linha
	 * @param coluna - numero da coluna
	 * @requires jogadaValida(linha, coluna) && jogada() > 1 && !terminado()
	 */	
	private void daTiro(int linha, int coluna){
		this.tabuleiro[(this.jogadorVez + 1) % NUMERO_JOGADORES].revela(linha, coluna);
		this.tirosDisparados++;
		if (this.tirosDisparados == TIROS) {
			passaVez();
			this.tirosDisparados = 0;
		}
	}

	/**
	 * Passa a vez para o proximo jogador.
	 */
	private void passaVez() {
		if (this.jogadorVez == 1)
			this.numJogada++;
		this.jogadorVez = (this.jogadorVez + 1) % NUMERO_JOGADORES;
	}

	/** 
	 * Verifica se o jogo esta terminado
	 * @return true se o jogo ja terminou, false caso contrario 
	 */	
	public boolean terminado() {	
		return this.numJogada >= 2 && 
		(this.tabuleiro[0].revelado() || this.tabuleiro[1].revelado());
	}

	/** 
	 * Indica o jogador vencedor 
	 * @return numero do jogador vencedor
	 * @requires terminado()
	 * @ensures \result ==1 || \result ==2
	 */
	public int vencedor() {
		return this.tabuleiro[0].revelado()? 2 : 1;
	}

	/**
	 * Representacao textual do estado da partida
	 * @return Devolve String representando estado da partida
	 */
	public String toString() {	
		
		StringBuilder sb = new StringBuilder();
		final String endLine = System.getProperty("line.separator");
		if (jogador() == 1) {
			sb.append("\nTabuleiro Jogador 1:\n");
			sb.append(this.tabuleiro[0].toString());
		} else {
			sb.append("\nTabuleiro Jogador 2:\n");
			sb.append(this.tabuleiro[1].toString());
		}
		if (terminado()) {
			sb.append(
				"Jogo terminado com vencedor: " + vencedor() + endLine);
		}
		else {
			sb.append("Jogada: " + jogada() + endLine + endLine);
			sb.append("Tem a vez o jogador " + jogador() + endLine);
			if (jogada() == 1) {
				sb.append("Submarinos escondidos: ");
				sb.append(this.submarinosEscondidos + endLine);
			}
			else {
				sb.append("Tiros disparados: " + 
					this.tirosDisparados + endLine);
			}
		}
		return sb.toString();
	}
}